#include <cpu.h>
#include <memory.h>
#include <debug.h>
#include <ports.h>

#define APIC_BASE 0xFEE00000

#define APIC_MSR_ENABLE (1<<11)

#define APIC_REG(r) (((uint32_t *)P2V(APIC_BASE + (r)))[0])

union ioapic_redirection {
    uint64_t val;
    struct {
        uint32_t l;
        uint32_t h;
    }__attribute__((packed));
    struct {
        uint8_t vector;
        uint8_t flags;
        uint8_t mask;
        uint32_t reserved;
        uint8_t target;
    }__attribute__((packed));
};

static uint32_t ioapic_read(int reg) {
    uint32_t volatile *ioregsel = P2V(ioapic.addr);
    uint32_t volatile *iowin = P2V(ioapic.addr + 0x10);
    *ioregsel = reg;
    return *iowin;
}
static void ioapic_write(int reg, uint32_t value) {
    uint32_t volatile *ioregsel = P2V(ioapic.addr);
    uint32_t volatile *iowin = P2V(ioapic.addr + 0x10);
    *ioregsel = reg;
    *iowin = value;
}

static uint64_t ioapic_read_redirection(int irq) {
    union ioapic_redirection retval;
    retval.l = ioapic_read(0x10 + 2*irq);
    retval.h = ioapic_read(0x11 + 2*irq);
    return retval.val;
}
static void ioapic_write_redirection(int irq, uint64_t val) {
    union ioapic_redirection value;
    value.val = val;
    ioapic_write(0x10 + 2*irq, value.l);
    ioapic_write(0x11 + 2*irq, value.h);
}

void ioapic_init() {
    vmm_set_page(kernel_P4, (uintptr_t)P2V(ioapic.addr), ioapic.addr, PAGE_PRESENT | PAGE_WRITE | PAGE_GLOBAL);

    union ioapic_redirection red;
    for(int i = 0; i < 24; i++) {
        red.val = ioapic_read_redirection(i);
        red.vector = IRQ_BASE+i;
        red.flags = 0x0;
        if(i == 1) // Mask the timer interrupt
            red.mask &= ~0x1;
        else // but unmask all others for testing
            red.mask |= 0x1;
        red.target = 0x0;
        ioapic_write_redirection(i, red.val);
    }
}

void apic_init() {
    // Enable local APIC
    write_msr(MSR_APIC_BASE, read_msr(MSR_APIC_BASE) | APIC_MSR_ENABLE);

    // Map apic offset into kernel memory
    vmm_set_page(kernel_P4, (uintptr_t)P2V(APIC_BASE), APIC_BASE, PAGE_PRESENT | PAGE_WRITE | PAGE_GLOBAL);

    APIC_REG(0xF0) = APIC_REG(0xF0) | (1<<8);
    APIC_REG(0xB0) = 0;
}